Bootstrap: docker
From: python:3.9-slim

%post
    # Install required libraries with specific versions
    pip install zarr==2.16.1 tifffile==2020.9.3 dask[array]==2021.3.0 numpy==1.26.4


%files
    convert_zarr_to_tiff.py /app/convert_zarr_to_tiff.py

%runscript
    # Run the script with the provided arguments
    exec python /app/convert_zarr_to_tiff.py "$@"
